
/*******************************************************************************
 * Copyright (C) 2017 - 2023, MINRES Technologies GmbH
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors:
 *       eyck@minres.com - initial API and implementation
 ******************************************************************************/

#include "cycle_estimate.h"
#include <yaml-cpp/yaml.h>

#include <iss/arch_if.h>
#include <util/logging.h>
#include <fstream>

using namespace std;

iss::plugin::cycle_estimate::cycle_estimate(string const& config_file_name)
: instr_if(nullptr)
, config_file_name(config_file_name)
{
}

iss::plugin::cycle_estimate::~cycle_estimate() {
}

bool iss::plugin::cycle_estimate::registration(const char* const version, vm_if& vm) {
    instr_if = vm.get_arch()->get_instrumentation_if();
    if(!instr_if) return false;
    const string  core_name = instr_if->core_type_name();
    if (config_file_name.length() > 0) {
        std::ifstream is(config_file_name);
        if (is.is_open()) {
            try {
                auto root = YAML::LoadAll(is);
                if(root.size()!=1) {
                    LOG(ERR) << "Too many root nodes in YAML file " << config_file_name;
                }
                for (auto p : root[0]) {
                    auto isa_subset = p.first;
                    auto instructions = p.second;
                    for (auto const& instr : instructions) {
                        instr_desc res;
                        res.is_branch = instr.second["branch"].as<bool>();
                        auto delay = instr.second["delay"];
                        if(delay.IsSequence()) {
                            res.not_taken = delay[0].as<uint64_t>();
                            res.taken = delay[1].as<uint64_t>();
                        } else {
                            res.not_taken = delay.as<uint64_t>();
                            res.taken = res.not_taken;
                        }
                        delays.push_back(std::move(res));
                    }
                }
            } catch (YAML::ParserException &e) {
                LOG(ERR) << "Could not parse input file " << config_file_name << ", reason: " << e.what();
                return false;
            }
        } else {
            LOG(ERR) << "Could not open input file " << config_file_name;
            return false;
        }
    }
    return true;

}

void iss::plugin::cycle_estimate::callback(instr_info_t instr_info) {
    assert(instr_if && "No instrumentation interface available but callback executed");
    auto entry = delays[instr_info.instr_id];
    bool taken = instr_if->is_branch_taken();
    if (taken && (entry.taken > 1))
        instr_if->update_last_instr_cycles(entry.taken);
    else if (entry.not_taken > 1)
        instr_if->update_last_instr_cycles(entry.not_taken);
}
